import os.path
import time
from urllib.parse import urlparse

import logging
logger = logging.getLogger(__name__)

import requests


class CitationSourceBase:

    def __init__(self, override_options, kwargs, default_options):
        super().__init__()

        self.options = dict(default_options)
        self.options.update(kwargs)
        self.options.update(override_options)

        self.chunk_size = self.options.get('chunk_size', 512)
        self.chunk_query_delay_ms = self.options.get('chunk_query_delay_ms', 1000)

        self.cite_prefix = self.options['cite_prefix']

        self.chains_to_sources = self.options.get('chains_to_sources', [])
        self.source_name = self.options.get('source_name', '<unknown source>')

        self.requests_session = None
        if self.options.get('use_requests', False):
            if 'requests_session' in self.options:
                self.requests_session = self.options['requests_session']
            else:
                self.requests_session = requests.Session()

        self.cite_key_list = None

        self.citation_manager = None

    def set_citation_manager(self, citation_manager):
        self.citation_manager = citation_manager


    def retrieve_citations(self, cite_key_list):

        self.cite_key_list = cite_key_list

        self.source_initialize_run()

        total_retrieved = 0
        remaining_keys = list(cite_key_list)

        last_chunk_query_monotonic_s = None

        if len(remaining_keys):
            logger.info(f"{self.source_name}: there are "
                        f"{len(remaining_keys)} citation(s) to query")
        else:
            logger.debug(f"{self.source_name}: there are no citations to query")

        while len(remaining_keys) > 0:

            # if applicable, wait before another chunk query call
            if (self.chunk_query_delay_ms and last_chunk_query_monotonic_s is not None):
                dt_s = (time.monotonic() - last_chunk_query_monotonic_s)
                dt_s_to_wait = (self.chunk_query_delay_ms/1000) - dt_s
                if dt_s_to_wait > 0:
                    time.sleep(dt_s_to_wait)

            # create a chunk of IDs to query
            chunk_ids, remaining_keys = \
                remaining_keys[:self.chunk_size], remaining_keys[self.chunk_size:]

            last_chunk_query_monotonic_s = time.monotonic()
            self.retrieve_chunk(chunk_ids)
            
            # bookkeeping & logging
            total_retrieved += len(chunk_ids)
            logger.info(f"{self.source_name}: {total_retrieved}/{len(cite_key_list)}")

        self.source_finalize_run()


    # helper for fetching URLs
    def fetch_url(self, url, binary=False, json=False, cwd='', **kwargs):

        if url is None:
            raise ValueError(f"fetch_url(): url is None!")

        logger.debug(f"Fetching URL ‘{url}’ ...")

        p = urlparse(url)

        if not p.scheme or p.scheme == 'file':
            # Read a local file.
            if binary:
                open_args, open_kwargs = ('rb',), {}
            else:
                open_args, open_kwargs = ('r',), { 'encoding': 'utf-8' }
            fname = os.path.join(cwd, p.path)
            with open(fname, *open_args, **open_kwargs) as f:
                return f.read()

        req_kwargs = {}

        method = kwargs.pop('method', 'get')
        reqfn = getattr(self.requests_session, method)

        if method == 'post':
            req_kwargs['data'] = kwargs.pop('body', None)

        # e.g., headers etc.
        req_kwargs.update(kwargs)

        # fire!
        r = reqfn(url)

        if not r.ok:
            # raise errors
            r.raise_for_status()

        if binary:
            return r.content

        r.encoding = 'utf-8'

        if json:
            return r.json()

        return r.text



    # -------------

    # can be reimplemented

    def source_initialize_run(self):
        pass

    def source_finalize_run(self):
        pass


    # must be reimplemented

    def retrieve_chunk(self, chunk_keys):
        raise RuntimeError("The method retrieve_chunk() must be reimplemented by subclasses!")



import re

import logging
logger = logging.getLogger(__name__)

import arxiv
import yaml

from .base import CitationSourceBase


_rx_arxiv_id_from_entryid = re.compile(
    # note: <arxivid> does NOT include the version number
    r'https?://arxiv.org/abs/(?P<arxivid>.*?)(?P<versionnumstr>v(?P<versionnum>\d+))?$',
    flags=re.IGNORECASE
)

_rx_ver = re.compile(r'v\d+$')

class CitationSourceArxiv(CitationSourceBase):

    def __init__(self, **kwargs):

        self.chain_to_doi = kwargs.get('chain_to_doi', True)

        override_options = {
            'chains_to_sources': ['doi'] if self.chain_to_doi else [],
            'source_name': 'ArXiv API citation info source',
        }
        default_options = {
            'chunk_query_delay_ms': 3100,
            'cite_prefix': 'arxiv',
            'use_requests': True,
        }

        super().__init__(
            override_options,
            kwargs,
            default_options,
        )

        self.override_arxiv_dois_file = self.options.get('override_arxiv_dois_file', None)
        self.override_arxiv_dois = self.options.get('override_arxiv_dois', {})

        # silence some arxiv.arxiv messages
        if not self.options.get('keep_arxiv_arxiv_logging_info_output', False):
            logging.getLogger('arxiv.arxiv').setLevel(level=logging.WARNING)


    def source_initialize_run(self):

        self._arxiv_to_doi_override = {}
        if self.override_arxiv_dois_file:
            content = self.fetch_url(self.override_arxiv_dois_file)
            data = yaml.safe_load(content)
            self._arxiv_to_doi_override.update(data)

        self._arxiv_to_doi_override.update(self.override_arxiv_dois)

        self.data_for_versionless_arxivid = {
            arxivid: []
            for arxivid in self.cite_key_list
            if not _rx_ver.search(arxivid)
        }
    
    def retrieve_chunk(self, arxivid_list):

        if not arxivid_list:
            return {}

        #
        # fetch meta-info from the arxiv for all encountered arXiv IDs, and
        # build the associated citation text endnotes.
        #
        big_slow_client = arxiv.Client(
            page_size=len(arxivid_list) + 100,
            delay_seconds=4,
            num_retries=5
        )
        searchobj = arxiv.Search(
            id_list=arxivid_list,
        )
        for result in big_slow_client.results(searchobj):
            #
            # build citation from the arxiv meta-information
            #

            m = _rx_arxiv_id_from_entryid.match(result.entry_id)
            if m is None:
                logger.warning(f"Unable to parse arXiv ID from {result.entry_id!r}")
                continue

            arxivid = m.group('arxivid').lower()
            versionnum = m.group('versionnum')
            arxivver = int(versionnum) if versionnum else None

            logger.debug(
                f"Processing received information for ‘{arxivid}’ (got v{arxivver})"
            )

            doi = None
            if result.doi is not None and result.doi:
                doi = str(result.doi)

            # override the given DOI in exceptional cases where it might be
            # missing, incomplete, or incorrect:
            if arxivid in self._arxiv_to_doi_override:
                override_doi = self._arxiv_to_doi_override[arxivid]
                logger.debug(f"Overriding reported DOI (‘{doi}’) for arXiv "
                             f"item ‘{arxivid}’ with manually specified DOI "
                             f"‘{override_doi}’ given in citation hints")
                doi = override_doi

            # construct CSL-JSON entry from the retrieved metadata
            citeprocjsond = {
                'type': 'article-journal',
                'title': result.title,
                'author': [
                    {'name': a.name}
                    for a in result.authors
                ],
                'published': {
                    'date-parts': [[
                        result.published.year,
                        result.published.month,
                        result.published.day,
                    ]],
                },
                'doi': doi,
                'arxivid': arxivid,
                'arxiv_version_number': arxivver,
            }
            
            if arxivid in self.cite_key_list:
                self.citation_manager.store_citation(self.cite_prefix, arxivid, citeprocjsond)

            if arxivid in self.data_for_versionless_arxivid:
                self.data_for_versionless_arxivid[arxivid].append( citeprocjsond )

        return

    def source_finalize_run(self):

        # all the queries for ids ran, now figure out which versions to link to
        # when a non-versioned key is provided

        for arxivid, versionslist in self.data_for_versionless_arxivid.items():

            if not versionslist:
                logger.error(
                    f"No arXiv data received for arXiv id ‘{arxivid}’, what happened?!?"
                )
                raise ValueError(f"No arXiv data found for ‘{arxivid}’")

            best = None
            for current in versionslist:
                if best is None and current is not None:
                    best = current
                    continue
                if current is None:
                    continue
                # "arxiv_version_number === null" only happens if arXiv API
                # responded with an article whose ID didn't have a version
                # number -> use that one directly as it is the answer to our
                # query!
                if best['arxiv_version_number'] is None:
                    continue
                if current['arxiv_version_number'] is None:
                    best = current
                    continue
                if current['arxiv_version_number'] >= best['arxiv_version_number']:
                    best = current
                    continue

            self._store_citation(arxivid, best)

    def _store_citation(self, arxivid, csljsondata):
        doi = csljsondata.get('doi', None)
        if doi and self.chain_to_doi:
            # chain to DOI
            self.citation_manager.store_citation_chained(
                self.cite_prefix, arxivid,
                'doi', doi,
                {
                    'arxivid': arxivid,
                },
            )
        else:
            self.citation_manager.store_citation(self.cite_prefix, arxivid, csljsondata)




# for when using shorthane naming
CitationSourceClass = CitationSourceArxiv



import re
import os.path
import sys
import json
import yaml

import logging
logger = logging.getLogger(__name__)


from pylatexenc.latexnodes import LatexWalkerError

from .base import CitationSourceBase


_vars = {
    'jobname': lambda doc: doc.metadata['jobname'],
}

_rx_vars = re.compile(r'\$\{(' + "|".join(re.escape(v) for v in _vars) + ')\}')

def _replace_vars(x, doc):
    return _rx_vars.sub(
        lambda m: _vars[m.group(1)] (doc),
        x
    )


class CitationSourceBibliographyFile(CitationSourceBase):

    def __init__(self, bibliography_file=None, **kwargs):

        override_options = {
            'chains_to_sources': [],
            'source_name': 'Bibliography file(s) citation info source',
            'chunk_size': sys.maxsize,
            'chunk_query_delay_ms': 0,
        }
        default_options = {
            'cite_prefix': 'bib',
            'use-requests': True, # to retrieve possibly remote bib files
        }

        super().__init__(
            override_options,
            kwargs,
            default_options,
        )

        doc = kwargs['doc']

        if doc.metadata and 'filepath' in doc.metadata:
            self.cwd = doc.metadata['filepath']['dirname']
        else:
            self.cwd = ''

        if bibliography_file is None:
            bibliography_files = []

            # specified manually in metadata
            if doc.metadata and 'bibliography' in doc.metadata:
                meta_bibliography_files = doc.metadata['bibliography']
                if isinstance(meta_bibliography_files, str):
                    meta_bibliography_files = [ meta_bibliography_files ]
                bibliography_files = bibliography_files + meta_bibliography_files

            # derived from jobname
            jobnamebibfile = doc.metadata['jobname'] + '.bib.json'
            if os.path.exists(jobnamebibfile):
                bibliography_files = bibliography_files + [ jobnamebibfile ]

        elif isinstance(bibliography_file, str):

            bibliography_files = [ bibliography_file ]

        else:

            bibliography_files = bibliography_file


        bibliography_files = [ _replace_vars(b, doc)
                               for b in bibliography_files ]

        self.bibliography_files = bibliography_files

        logger.debug(f"bib file citation source ({self.cite_prefix=}), "
                     f"{self.bibliography_files=}")
            
        self.bibliography_data = {}

    def source_initialize_run(self):
        
        for bibfile in self.bibliography_files:
            logger.debug(f"Loading bibliography ‘{bibfile}’ ...")
            bibdata = self.fetch_url(bibfile, cwd=self.cwd)
            if bibfile.endswith('.json'):
                bibdatajson = json.loads(bibdata)
            elif bibfile.endswith( ('.yml', '.yaml') ):
                bibdatajson = yaml.safe_load(bibdata)
            else:
                raise LatexWalkerError(
                    f"Unknown bibliography format: ‘{bibfile}’ (expected "
                    f"CSL-JSON or CSL-YAML)"
                )
                
            if isinstance(bibdatajson, list):
                bibdatajson = {
                    obj['id']: obj
                    for obj in bibdatajson
                }
            self.bibliography_data.update(bibdatajson)

    def retrieve_chunk(self, chunk_keys):

        for key in chunk_keys:
            if key not in self.bibliography_data:
                raise LatexWalkerError(
                    f"Bibliography key {key} was not found in bibliography file(s) "
                    + ', '.join([f'‘{b}’' for b in self.bibliography_files])
                )

            self.citation_manager.store_citation(
                self.cite_prefix, key, self.bibliography_data[key]
            )

        return



# for when using shorthane naming
CitationSourceClass = CitationSourceBibliographyFile



import sys

from .base import CitationSourceBase


class CitationSourceManual(CitationSourceBase):

    def __init__(self, **kwargs):

        override_options = {
            'chains_to_sources': [],
            'source_name': 'Manual citation info source',
            'chunk_size': sys.maxsize,
            'chunk_query_delay_ms': 0,
        }
        default_options = {
            'cite_prefix': 'manual',
            'use_requests': False,
        }

        super().__init__(
            override_options,
            kwargs,
            default_options,
        )


    def retrieve_chunk(self, chunk_keys):

        for key in chunk_keys:
            self.citation_manager.store_citation(
                self.cite_prefix, key,
                {
                    '_formatted_flm_text': key, # well that was hard
                }
            )

        return



# for when using shorthane naming
CitationSourceClass = CitationSourceManual




from pylatexenc.latexnodes.nodes import LatexNodesVisitor


class CitationsScanner(LatexNodesVisitor):
    def __init__(self):
        super().__init__()
        self.encountered_citations = []

    def get_encountered_citations(self):
        return self.encountered_citations

    # ---

    def visit_macro_node(self, node, **kwargs):
        if hasattr(node, 'flmarg_cite_items'):
            # it's a citation node with citations to track
            for cite_item in node.flmarg_cite_items:
                cite_prefix, cite_key, cite_extra = \
                    cite_item['prefix'], cite_item['key'], cite_item['extra']
                self.encountered_citations.append(
                    dict(
                        cite_prefix=cite_prefix,
                        cite_key=cite_key,
                        encountered_in=dict(
                            resource_info=node.latex_walker.resource_info,
                            what=f"{node.latex_walker.what} @ {node.latex_walker.pos_to_lineno_colno(node.pos)!r}",
                        )
                    )
                )

        super().visit_macro_node(node)



import re

from urllib.parse import quote as urlquote

import backoff
import requests

from .base import CitationSourceBase


class CitationSourceDoi(CitationSourceBase):

    def __init__(self, **kwargs):

        req_session = requests.Session()
        headers={
            'Accept': 'application/vnd.citationstyles.csl+json'
        }
        req_session.headers.update(headers)

        override_options = {
            'chains_to_sources': [],
            'source_name': 'DOI citation info source',
            'chunk_size': 1, # one DOI at a time
        }
        default_options = {
            'chunk_query_delay_ms': 1100,
            'cite_prefix': 'doi',
            'use_requests': True,
            'requests_session': req_session,
        }

        super().__init__(
            override_options,
            kwargs,
            default_options,
        )


    def retrieve_chunk(self, chunk_keys):

        if ( len(chunk_keys) != 1 ):
            raise RuntimeError(
                f"doi retrieve_chunk(): can only query one DOI at a time, "
                f"requested {chunk_keys!r}"
            )

        doi = chunk_keys[0].strip()

        self.citation_manager.store_citation(
            self.cite_prefix, doi, _get_doi_citeproc_json_object(doi, self)
        )





# utility to get the citeproc-json for a given DOI, with backoff
# (cf. https://www.crossref.org/documentation/retrieve-metadata/rest-api/tips-for-using-the-crossref-rest-api/)

def _backoff_handler(details):
    logger.warning("Backing off {wait:0.1f} seconds after {tries} tries "
                   "calling function {target} with args {args} and kwargs "
                   "{kwargs}".format(**details), exc_info=True)

def _backoff_fatal_code(e):
    if not e.response:
        return True
    return e.response.status_code == 404

@backoff.on_exception(backoff.expo,
                      requests.exceptions.RequestException,
                      max_tries=12,
                      on_backoff=_backoff_handler,
                      giveup=_backoff_fatal_code)
def _get_doi_citeproc_json_object(doi, source):

    doi_escaped = urlquote(doi)

    url = f"https://doi.org/{doi_escaped}"

    return source.fetch_url(
        url,
        json=True,
        # Accept: header already set in source's request_session
    )


# for when using shorthane naming
CitationSourceClass = CitationSourceDoi





import re

_escape_chars = {
    '\\': r'\textbackslash',
    '%': r'\%',
    '#': r'\#',
    '&': r'\&',
    '$': r'\$',
    '{': r'\{',
    '}': r'\}',
}

def _multiple_replace(s, rep_dict):
    # https://stackoverflow.com/a/15448887/1694896
    pattern = re.compile("|".join([
        re.escape(k)
        for k in sorted(rep_dict.keys(),key=len,reverse=True)
    ]), flags=re.DOTALL)
    return pattern.sub(lambda x: rep_dict[x.group(0)], s)

def preformat(text):
    # if isinstance(text, list): # ???
    #     return [preformat(z) for z in text]
    text = str(text)

    s = _multiple_replace(text, _escape_chars)

    return s


class FLMCommandWrapper(str):
    cmd = None

    @classmethod
    def _wrap(cls, text):
        text = preformat(text)
        return f'\\{cls.cmd}{{{text}}}'

    def __new__(cls, text):
        return super(FLMCommandWrapper, cls).__new__(cls, cls._wrap(text))


class Italic(FLMCommandWrapper):
    cmd = 'textit'


class Oblique(Italic):
    pass


class Bold(FLMCommandWrapper):
    cmd = 'textbf'


class Light(FLMCommandWrapper):
    cmd = 'textit' # italic instead


class Underline(FLMCommandWrapper):
    cmd = 'textit' # italic instead


class Superscript(FLMCommandWrapper):
    cmd = 'textit' # italic instead


class Subscript(FLMCommandWrapper):
    cmd = 'textit' # italic instead


class SmallCaps(FLMCommandWrapper):
    cmd = 'textit' # italic instead





### FIXME: Merge feature config of imported configs don't work yet ... :/
### Need to fix this in flm.main.run().  Probably all $import's should be
### processed first, before merging anything.

flm_default_import_config = {
    'flm': {
        'features': {
            'flm_citations': {},
            # {
            #     'sources': [
            #         {
            #             '$defaults': True,
            #         },
            #     ],
            # },
        },
    },
}


from .feature import FeatureClass



import json
import importlib
import warnings
import copy
import os.path
import datetime
from urllib.parse import quote as urlquote

import logging
logger = logging.getLogger(__name__)

from pylatexenc.latexnodes import LatexWalkerError

from flm.feature.cite import (
    FeatureExternalPrefixedCitations,
)

import citeproc
import citeproc.source
import citeproc.source.json
from . import _cslformatter

from .flmcitationsscanner import CitationsScanner


def importclass(fullname):
    modname, classname = fullname.rsplit('.', maxsplit=1)
    mod = importlib.import_module(modname)
    return getattr(mod, classname)


_default_citation_sources_spec = [
    {
        'name': 'arxiv',
        'config': {},
    },
    {
        'name': 'doi',
        'config': {},
    },
    {
        'name': 'manual',
        'config': {},
    },
    {
        'name': 'bibliographyfile',
        'config': {},
    },
]

class FeatureCiteAuto(FeatureExternalPrefixedCitations):
    r"""
    .....

    Arguments:

    - `citation_sources` - a dictionary where keys are valid `cite_prefix`'s and
      where each value is an instance of `CitationSource` (see `citesources` module).

    - ... further arguments are passed on to
      `flm.feature.cite.FeatureExternalPrefixedCitations`.
    """

    add_arxiv_link = True
    add_doi_link = True
    add_url_link = 'only-if-no-other-link' # = only added if there's no arxiv or doi link

    # inherited!!
    #feature_name = 'citations'

    feature_default_config = {
        'sources': _default_citation_sources_spec,
    }

    class RenderManager(FeatureExternalPrefixedCitations.RenderManager):
        
        def get_citation_content_flm(self, cite_prefix, cite_key, resource_info):

            fdocmgr = self.feature_document_manager

            csljson = fdocmgr.get_citation_csljson(cite_prefix, cite_key)

            result = _generate_citation_flm_from_citeprocjsond(
                csljson,
                bib_csl_style=fdocmgr.bib_csl_style,
                what=str(resource_info),
                flm_environment=self.render_context.doc.environment,
                add_arxiv_link=self.feature.add_arxiv_link,
                add_doi_link=self.feature.add_doi_link,
                add_url_link=self.feature.add_url_link,
            )

            return result

    class DocumentManager(FeatureExternalPrefixedCitations.DocumentManager):

        def initialize(self):
            super().initialize()

            self.citation_sources = {}

            for citation_source_spec in self.feature.sources:

                cname = citation_source_spec['name']
                cconfig = citation_source_spec.get('config', {})

                if '.' not in cname:
                    cname = f"flm_citations.citesources.{cname}.CitationSourceClass"

                cconfig.update(doc=self.doc)

                TheClass = importclass(cname)
                thesource = TheClass(**cconfig)
                thesource.set_citation_manager(self)
                self.citation_sources[thesource.cite_prefix] = thesource

            bib_csl_style = self.feature.bib_csl_style
            if bib_csl_style is None:
                bib_csl_style = os.path.join(
                    os.path.dirname(__file__),
                    'american-physical-society-et-al--patched.csl'
                )
                #"harvard1"
            self.bib_csl_style = \
                citeproc.CitationStylesStyle(bib_csl_style, validate=False)

            self.citations_db = {
                cite_prefix: {}
                for cite_prefix in self.citation_sources.keys()
            }

            logger.debug(f"citation_sources are {self.citation_sources!r}")

            doc = self.doc
            if doc.metadata and 'filepath' in doc.metadata:
                self.cwd = doc.metadata['filepath']['dirname']
            else:
                self.cwd = ''

            self.cache_file_path = os.path.join(self.cwd, self.feature.cache_file)

            self.load_cache()

            self.new_chained_citations = None

        def load_cache(self):
            if os.path.exists(self.cache_file_path):
                #logger.debug(f"Loading cache file ‘{self.cache_file_path}’")
                try:
                    with open(self.cache_file_path, 'r', encoding='utf-8') as f:
                        self.citations_db.update(json.load(f))
                    # check for expired entries
                    now = datetime.datetime.now()
                    for cite_prefix, cite_key_dict in self.citations_db.items():
                        cite_keys_list = list(cite_key_dict.keys())
                        for cite_key in cite_keys_list:
                            dexpires = datetime.datetime.fromisoformat(
                                self.citations_db[cite_prefix][cite_key]['expires']
                            )
                            if dexpires < now:
                                del self.citations_db[cite_prefix][cite_key]

                    logger.debug("Loaded cache ‘%s’: %s", self.cache_file_path,
                                 ", ".join([
                                     f"{len(cprefixdb)} {cprefix}"
                                     for cprefix, cprefixdb in self.citations_db.items()
                                 ]))
                except Exception as e:
                   logger.warning(
                       f"Failure while loading cache file ‘{self.cache_file_path}’: "
                       f"{e}, ignoring ..."
                   )
                   return

        def save_cache(self):
            with open(self.cache_file_path, 'w', encoding='utf-8') as fw:
                json.dump(self.citations_db, fw)


        def get_citation_csljson(self, cite_prefix, cite_key):
            orig_cite_prefix, orig_cite_key = cite_prefix, cite_key
            set_properties_chain = {}
            while True:
                csljson = self.citations_db[cite_prefix][cite_key]['entry']
                if 'chained' not in csljson:
                    # make sure we have the correct ID set
                    origid = f"{orig_cite_prefix}:{orig_cite_key}"
                    if set_properties_chain:
                        csljson = dict(csljson, **set_properties_chain)
                    if csljson['id'] != origid:
                        return dict(csljson, id=origid)
                    return csljson

                # chained citation, follow chain
                cite_prefix = csljson['chained']['cite_prefix']
                cite_key = csljson['chained']['cite_key']
                set_properties_chain = dict(
                    csljson['chained']['set_properties'],
                    **set_properties_chain
                )

                # ... and repeat !

        def store_citation(self, cite_prefix, cite_key, csljson):

            if csljson.get('chained', None):
                new_cite_prefix = csljson['chained']['cite_prefix']
                new_cite_key = csljson['chained']['cite_key']

                if new_cite_prefix not in self.citation_sources:
                    raise LatexWalkerError(
                        f"No source registered for cite prefix ‘{new_cite_prefix}’ in "
                        f"chained citation retreival for ‘{cite_prefix}:{cite_key}’"
                    )

            cslentry = dict(csljson, id=f"{cite_prefix}:{cite_key}")

            self.citations_db[cite_prefix][cite_key] = {
                'entry': cslentry,
                'expires': (datetime.datetime.now()
                            + self.feature.cache_entry_duration_dt).isoformat()
            }

            # save cache at each store
            self.save_cache()

        def store_citation_chained(self, cite_prefix, cite_key,
                                   new_cite_prefix, new_cite_key, set_properties):

            self.store_citation(
                cite_prefix,
                cite_key,
                # special JSON entry to store in cache
                {
                    'chained': {
                        'cite_prefix': new_cite_prefix,
                        'cite_key': new_cite_key,
                        'set_properties': dict(set_properties),
                    }
                },
            )

            self.new_chained_citations.append( (cite_prefix, cite_key) )


        def flm_main_scan_fragment(self, fragment, document_parts_fragments=None, **kwargs):

            scanner = CitationsScanner()

            fragment.start_node_visitor(scanner)
            if document_parts_fragments:
                for frag in document_parts_fragments:
                    frag.start_node_visitor(scanner)

            retrieve_citation_keys_by_prefix = {
                cite_prefix: set()
                for cite_prefix in self.citation_sources.keys()
            }

            for c in scanner.get_encountered_citations():
                logger.debug(f"Found citation {c=!r}")

                if c['cite_prefix'] not in retrieve_citation_keys_by_prefix:
                    raise LatexWalkerError(
                        f"Invalid citation prefix ‘{c['cite_prefix']}’ in "
                        f"{c['encountered_in']['what']}"
                    )

                retrieve_citation_keys_by_prefix[c['cite_prefix']].add(c['cite_key'])

            while any(retrieve_citation_keys_by_prefix.values()):

                self.new_chained_citations = []

                for cite_prefix, cite_key_set in retrieve_citation_keys_by_prefix.items():

                    retrieve_cite_key_set = [
                        cite_key
                        for cite_key in cite_key_set
                        if cite_key not in self.citations_db[cite_prefix]
                    ]

                    logger.debug(f"Keys to retrieve: {cite_prefix} -> {retrieve_cite_key_set}")
                    self.citation_sources[cite_prefix].retrieve_citations(
                        retrieve_cite_key_set
                    )

                #logger.debug(f"At this point, {self.citations_db = }")

                retrieve_citation_keys_by_prefix = {
                    cite_prefix: set()
                    for cite_prefix in self.citation_sources.keys()
                }

                # check if there are chained citations that we need to retrieve as well
                for cite_prefix, cite_key in self.new_chained_citations:
                    csljson = self.citations_db[cite_prefix][cite_key]['entry']
                    chained = csljson['chained']
                    chained_cite_prefix, chained_cite_key = \
                        chained['cite_prefix'], chained['cite_key']
                    if chained_cite_key not in self.citations_db[chained_cite_prefix]:
                        # add this one for retrieval
                        retrieve_citation_keys_by_prefix[chained_cite_prefix].add(
                            chained_cite_key
                        )
                        #logger.debug(f"Adding ‘{chained_cite_prefix}:{chained_cite_key}’ "
                        #             f"for retrieval")



    def __init__(self,
                 sources=None,
                 bib_csl_style=None,
                 cache_file='.flm-citations.cache.json',
                 cache_entry_duration_dt=datetime.timedelta(days=30),
                 **kwargs):

        super().__init__(external_citations_providers=None, **kwargs)

        if sources is None:
            sources = _default_citation_sources_spec
        self.sources = sources

        self.bib_csl_style = bib_csl_style

        self.cache_file = cache_file
        self.cache_entry_duration_dt = cache_entry_duration_dt



def _generate_citation_flm_from_citeprocjsond(
        citeprocjsond, bib_csl_style, what, flm_environment, *,
        add_arxiv_link, add_doi_link, add_url_link,
):

    if '_formatted_flm_text' in citeprocjsond:
        # work is already done for us -- go!
        return flm_environment.make_fragment(
            citeprocjsond['_formatted_flm_text'],
            what=what,
            standalone_mode=True,
        )

    with warnings.catch_warnings():
        if hasattr(citeproc.source, 'MissingArgumentWarning'):
            # my patched version
            warnings.simplefilter('ignore', citeproc.source.MissingArgumentWarning)
            warnings.simplefilter('ignore', citeproc.source.UnsupportedArgumentWarning)
        else:
            # until citeproc-py merges my PR
            warnings.simplefilter('ignore', UserWarning)

        citekey = citeprocjsond['id']

        logger.debug(f"Creating citation for entry ‘{citekey}’")

        # patch JSON for limitations of citeproc-py (?)
        #
        # E.g. for authors with 'name': ... instead of 'given': and 'family':

        if 'author' in citeprocjsond:
            citeprocjsond = copy.copy(citeprocjsond)
            for author in citeprocjsond['author']:
                if 'name' in author and 'family' not in author and 'given' not in author:
                    author['family'] = author['name']
                    del author['name']

        # explore the citeprocjsond tree and make sure that all strings are
        # valid FLM markup
        def _sanitize(d):
            if isinstance(d, dict):
                for k in d.keys():
                    d[k] = _sanitize(d[k])
                return d
            elif isinstance(d, list):
                for j, val in enumerate(d):
                    d[j] = _sanitize(val)
                return d
            else:
                try:
                    # try compiling the given value, suppressing warnings
                    flm_environment.make_fragment(
                        str(d),
                        standalone_mode=True,
                        silent=True
                    )
                except Exception as e:
                    logger.debug(
                        f"Encountered invalid FLM string {d!r} when "
                        f"composing citation: {e}"
                    )
                    return r'\begin{verbatimtext}' + str(d) + r'\end{verbatimtext}'
                return d

        #
        # Sanitizing the entire JSON object (which often includes the abstract,
        # etc.) is completely overkill.  So we first try to generate the entry
        # without sanitizing, and if it fails, we sanitize.
        #
        #_sanitize(citeprocjsond)

        def _gen_entry(citeprocjsond):
            bib_source = citeproc.source.json.CiteProcJSON([citeprocjsond])
            bibliography = citeproc.CitationStylesBibliography(bib_csl_style, bib_source,
                                                               _cslformatter)

            citation1 = citeproc.Citation([citeproc.CitationItem(citeprocjsond['id'])])
            bibliography.register(citation1)
            bibliography_items = [str(item) for item in bibliography.bibliography()]
            assert len(bibliography_items) == 1
            result = bibliography_items[0]

            arxivid = citeprocjsond.get('arxivid', None)
            doi = citeprocjsond.get('doi', None) or citeprocjsond.get('DOI', None)
            url = citeprocjsond.get('URL', None)
            if doi and add_doi_link:
                doiurl = 'https://doi.org/'+urlquote(doi)
                result += ' \href{'+doiurl+'}{DOI}'
            if arxivid and add_arxiv_link:
                result += ' \href{https://arxiv.org/abs/'+arxivid+'}{'+arxivid+'}'
            if url and (add_url_link is True or
                        (add_url_link == 'only-if-no-other-link'
                         and not arxivid and not doi)):
                result += ' \href{'+url+'}{URL}'

            return result

        try:
            logger.debug(f"Attempting to generate entry for {citekey}...")
            return flm_environment.make_fragment(
                _gen_entry(citeprocjsond),
                what=what,
                standalone_mode=True,
                silent=True # don't report errors on logger
            )
        except Exception:
            logger.debug(f"Error while forming citation entry for {citekey}, trying "
                         f"again with FLM sanitization on")

        _sanitize(citeprocjsond)
        try:
            return flm_environment.make_fragment(
                _gen_entry(citeprocjsond),
                standalone_mode=True,
                what=what
            )
        except Exception as e:
            logger.critical(f"EXCEPTION!! {e!r}")
            raise


# ------------------------------------------------

FeatureClass = FeatureCiteAuto
